"""Defines the WildfireEnv class, which simulates dynamics of unmanned aerial vehicles (UAVs) fighting a spreading wildfire
"""

from collections import OrderedDict
import random
from gym.spaces import Box, Dict, Discrete
import numpy as np
from wildfire_environment.multigrid import MultiGridEnv
from wildfire_environment.core.world import WildfireWorld
from wildfire_environment.core.agent import WildfireActions, Agent
from wildfire_environment.core.object import Tree
from wildfire_environment.core.grid import Grid
from wildfire_environment.core.constants import (
    STATE_TO_IDX_WILDFIRE,
    TILE_PIXELS,
    STATE_IDX_TO_COLOR_WILDFIRE,
    COLORS,
)
from wildfire_environment.utils.window import Window
from wildfire_environment.utils.misc import (
    render_agent_tiles,
    get_initial_fire_coordinates,
)


class WildfireEnv(MultiGridEnv):
    """Grid environment which simulates dynamics of unmanned aerial vehicles (UAVs) fighting a spreading wildfire"""

    def __init__(
        self,
        alpha=0.05,
        beta=0.99,
        delta_beta=0,
        size=17,
        num_agents=2,
        agent_start_positions=((1, 1), (15, 15)),
        agent_colors=("red", "blue"),
        agent_groups=None,
        agent_view_size=10,
        initial_fire_size=1,
        max_steps=100,
        partial_obs=False,
        actions_set=WildfireActions,
        render_mode="rgb_array",
        render_selfish_region_boundaries=False,
        cooperative_reward=False,
        log_selfish_region_metrics=False,
        selfish_region_xmin=None,
        selfish_region_xmax=None,
        selfish_region_ymin=None,
        selfish_region_ymax=None,
    ):
        """Create a WildfireEnv environment

        Parameters
        ----------
        alpha : float, optional
            parameter for the wildfire dynamics model, by default 0.05
        beta : float, optional
            parameter for the wildfire dynamics model, by default 0.99
        delta_beta : float, optional
            parameter for the wildfire dynamics model, by default 0
        size : int, optional
            side of the square gridworld, by default 17
        num_agents : int, optional
            number of UAV agents, by default 2
        agent_start_positions : tuple[tuple[int,int]], optional
            tuple of tuples containing the start positions of the agents, in order of agent index. By default ((1, 1), (15, 15))
        agent_colors : tuple[str,str], optional
            tuple of strings of color names of all agents (or groups if agents are grouped) in order of increasing index. All agents in a group have the same color. The strings should be keys in the COLORS dictionary in constants.py. Only applicable if cooperative_reward is False. Fully cooperative agents have light_blue color by default. By default self-interested agents have red and blue colors
        agent_groups : tuple[tuple], optional
            tuple of tuples containing the indices (in ascending order) of agents in each group. Only applicable if cooperative_reward is False. By default None
        agent_view_size : int, optional
            side of the square region visible to an agent with partial observability, by default 10. Only applicable if partial_obs is True
        initial_fire_size : int, optional
            side of the square shaped initial fire region, by default 1
        max_steps : int, optional
            maximum number of steps in an episode, by default 100
        partial_obs : bool, optional
            whether agents have partial observability, by default False
        actions_set : WildfireActions, optional
            action space of the agents. All agents have the same action space. By default WildfireActions.
        render_mode : str, optional
            mode of rendering the environment, by default "rgb_array"
        render_selfish_region_boundaries : bool, optional
            whether to render boundaries of selfish regions, by default False
        cooperative_reward : bool, optional
            whether the agents use a cooperative reward, by default False. If True, the agents are fully cooperative and receive the same reward.
        log_selfish_region_metrics : bool, optional
            whether to log metrics related to trees in selfish regions, by default False
        selfish_region_xmin : list, optional
            list containing x-coordinates of the left boundaries of the regions of selfish interest for the agents (or groups if the agents are grouped. All agents in a group have same region of selfish interest). Regions of selfish interest are rectangular. List elements are in order of agent (or group) indices. Only applicable if cooperative_reward is False. By default None.
        selfish_region_xmax : list, optional
            list containing x-coordinates of the right boundaries of the regions of selfish interest for the agents (or groups if the agents are grouped. All agents in a group have same region of selfish interest). Regions of selfish interest are rectangular. List elements are in order of agent (or group) indices. Only applicable if cooperative_reward is False. By default None.
        selfish_region_ymin : list, optional
            list containing y-coordinates of the top boundaries of the regions of selfish interest for the agents (or groups if the agents are grouped. All agents in a group have same region of selfish interest). Regions of selfish interest are rectangular. List elements are in order of agent (or group) indices. Only applicable if cooperative_reward is False. By default None.
        selfish_region_ymax : list, optional
            list containing y-coordinates of the bottom boundaries of the regions of selfish interest for the agents (or groups if the agents are grouped. All agents in a group have same region of selfish interest). Regions of selfish interest are rectangular. List elements are in order of agent (or group) indices. Only applicable if cooperative_reward is False. By default None.
        """
        self.alpha = alpha
        self.beta = beta
        self.delta_beta = delta_beta
        self.num_agents = num_agents
        self.agent_start_positions = agent_start_positions
        self.agent_colors = agent_colors
        self.agent_groups = agent_groups
        if agent_groups:
            self.idx_to_group = {}
            for i, group in enumerate(agent_groups):
                for agent_index in group:
                    self.idx_to_group[agent_index] = i
        # observation vector of each agent is concatenation of obs_depth number of one-hot encodings, see paper for details. len(STATE_IDX_TO_COLOR_WILDFIRE) = the number of tree states
        self.obs_depth = self.num_agents + len(STATE_IDX_TO_COLOR_WILDFIRE)
        self.max_steps = max_steps
        self.world = WildfireWorld
        self.grid_size = size
        self.grid_size_without_walls = size - 2
        self.initial_fire_size = initial_fire_size
        self.burnt_trees = 0
        self.unburnt_trees = []
        self.trees_on_fire = 0
        self.cooperative_reward = cooperative_reward
        self.render_selfish_region_boundaries = render_selfish_region_boundaries
        self.log_selfish_region_metrics = log_selfish_region_metrics
        if self.log_selfish_region_metrics:
            # initialize attributes for logging metrics related to trees in selfish regions
            self.selfish_xmin = np.array(selfish_region_xmin)
            self.selfish_xmax = np.array(selfish_region_xmax)
            self.selfish_ymin = np.array(selfish_region_ymin)
            self.selfish_ymax = np.array(selfish_region_ymax)
            self.selfish_region_trees_on_fire = np.zeros(len(self.selfish_xmin))
            self.selfish_region_burnt_trees = np.zeros(len(self.selfish_xmin))
            self.selfish_region_size = (
                self.selfish_xmax
                - self.selfish_xmin
                + np.ones(len(selfish_region_xmin))
            ) * (
                self.selfish_ymax
                - self.selfish_ymin
                + np.ones(len(selfish_region_ymin))
            )

        if self.cooperative_reward:
            # initialize cooperative agents
            agents = [
                Agent(
                    world=self.world,
                    index=i,
                    view_size=agent_view_size,
                    actions=actions_set,
                    color="light_blue",
                )
                for i in range(self.num_agents)
            ]
        elif self.agent_groups:
            # initialize self-interested agents with different colors
            agents = [
                Agent(
                    world=self.world,
                    index=i,
                    view_size=agent_view_size,
                    actions=actions_set,
                    color=self.agent_colors[self.idx_to_group[i]],
                )
                for i in range(self.num_agents)
            ]
        else:
            # initialize self-interested agents with different colors
            agents = [
                Agent(
                    world=self.world,
                    index=i,
                    view_size=agent_view_size,
                    actions=actions_set,
                    color=self.agent_colors[i],
                )
                for i in range(self.num_agents)
            ]

        super().__init__(
            agents=agents,
            grid_size=size,
            max_steps=max_steps,
            partial_obs=partial_obs,
            agent_view_size=agent_view_size,
            actions_set=actions_set,
            world=self.world,
            render_mode=render_mode,
        )
        self.helper_grid = None
        self.observation_space = self._set_observation_space()
        self.action_space = Dict(
            {f"{a.index}": Discrete(len(self.actions)) for a in self.agents}
        )

    def _set_observation_space(self) -> Dict:
        """Set the observation space for each agent in the environment. All agents possess the same observation space

        Returns
        -------
        observation_space : dict
            dictionary where each key is an agent's index and the value is the observation space
            for that agent
        """
        # observation vector of agent is the concatenation of obs_depth number of one-hot encodings where each encoding has grid_size_without_walls number of elements valued either 0 or 1. Additionally, the observation vector contains the normalized time step at the end
        low = np.full(self.obs_depth * ((self.grid_size_without_walls + 1) ** 2) + 1, 0)
        high = np.full(
            self.obs_depth * ((self.grid_size_without_walls + 1) ** 2) + 1, 1
        )
        observation_space = Dict(
            {
                f"{a.index}": Box(
                    low=low,
                    high=high,
                    dtype=np.float32,
                )
                for a in self.agents
            }
        )

        return observation_space

    def _gen_grid(self, width, height, state=None):
        """Generate the grid for the environment

        Parameters
        ----------
        width : int
            width of the grid
        height : int
            height of the grid
        state : ndarray, optional
            specifies the initial state of the environment, by default None.
            If none, it is chosen uniformly at random from the assumed initial state distribution
        """
        self.grid = Grid(width, height, self.world)

        # generate the walls of the grid
        self.grid.horz_wall(0, 0)
        self.grid.horz_wall(0, height - 1)
        self.grid.vert_wall(0, 0)
        self.grid.vert_wall(width - 1, 0)

        agent_start_pos = []
        if state is not None:
            # store positions of agents and trees on fire as per the specified initial state.
            state = state[:-1].reshape(
                (
                    self.obs_depth + 1,
                    self.grid_size,
                    self.grid_size,
                ),
            )
            initial_fire = []
            for i in range(self.grid_size):
                for j in range(self.grid_size):
                    if state[1, j, i]:
                        # i and j are swapped because the y-coordinate specifies the row, while the x-coordinate specifies the column.
                        initial_fire.append((i, j))
                    for o in self.agents:
                        index = o.index
                        if state[len(STATE_IDX_TO_COLOR_WILDFIRE) + 1 + index, j, i]:
                            # i and j are swapped because the y-coordinate specifies the row, while the x-coordinate specifies the column.
                            agent_start_pos.append((i, j))

        else:
            # choose location of initial fire uniformly at random
            if self.initial_fire_size % 2 == 0:
                # for even sized initial fires, choose location of top left corner of fire region uniformly at random
                top_left_corner = (
                    random.randint(
                        1,
                        self.grid_size_without_walls - (self.initial_fire_size),
                    ),
                    random.randint(
                        1,
                        self.grid_size_without_walls - (self.initial_fire_size),
                    ),
                )
                initial_fire = get_initial_fire_coordinates(
                    *top_left_corner,
                    self.grid_size_without_walls,
                    self.initial_fire_size,
                )
            else:
                # for odd sized initial fires, choose location of center of fire region uniformly at random
                fire_square_center = (
                    random.randint(
                        1 + ((self.initial_fire_size - 1) / 2),
                        self.grid_size_without_walls
                        - ((self.initial_fire_size - 1) / 2),
                    ),
                    random.randint(
                        1 + ((self.initial_fire_size - 1) / 2),
                        self.grid_size_without_walls
                        - ((self.initial_fire_size - 1) / 2),
                    ),
                )
                initial_fire = get_initial_fire_coordinates(
                    *fire_square_center,
                    self.grid_size_without_walls,
                    self.initial_fire_size,
                )
            # agent_start_pos is specified during environment initialization
            agent_start_pos = self.agent_start_positions

        for pos in initial_fire:
            region = "common"
            if self.log_selfish_region_metrics:
                # update count of trees on fire in selfish regions
                if self.agent_groups:
                    for i, _ in enumerate(self.agent_groups):
                        if self.in_selfish_region(pos[0], pos[1], i):
                            # selfish region is identified by the lowest index among indices of the corresponding group of selfish agents
                            region = f"{i}"
                            self.selfish_region_trees_on_fire[i] += 1
                            break
                else:
                    for a in self.agents:
                        if self.in_selfish_region(pos[0], pos[1], a.index):
                            # selfish region is identified by the lowest index among indices of the corresponding selfish agent
                            region = f"{a.index}"
                            self.selfish_region_trees_on_fire[a.index] += 1
                            break
            # insert tree on fire in grid
            self.put_obj(
                Tree(self.world, STATE_TO_IDX_WILDFIRE["on fire"], region=region),
                int(pos[0]),
                int(pos[1]),
            )

        # update counts of trees on fire and healthy trees
        self.trees_on_fire += self.initial_fire_size**2
        num_healthy_trees = self.grid_size_without_walls**2 - len(initial_fire)

        # insert healthy tree in grid
        for _ in range(num_healthy_trees):
            tree_obj = Tree(self.world, STATE_TO_IDX_WILDFIRE["healthy"])
            self.place_obj(tree_obj)
            if self.log_selfish_region_metrics:
                # check if tree is in a selfish region, and update region attribute of tree if it is
                if self.agent_groups:
                    for i, _ in enumerate(self.agent_groups):
                        if self.in_selfish_region(
                            *(tree_obj.pos), i  # pylint: disable=not-an-iterable
                        ):
                            tree_obj.region = f"{i}"
                            break
                else:
                    for a in self.agents:
                        if self.in_selfish_region(
                            *(tree_obj.pos), a.index  # pylint: disable=not-an-iterable
                        ):
                            tree_obj.region = f"{a.index}"
                            break

        # helper grid is a work around for grid being unable to store multiple objects at a single cell. It does not contain agents
        self.helper_grid = self.grid.copy()

        # create list of unburnt trees. initial state does not have burnt trees.
        for c in self.helper_grid.grid:
            if c is not None and c.type == "tree":
                self.unburnt_trees.append(c)

        # insert agents in grid
        for i, a in enumerate(self.agents):
            self.place_agent(a, pos=agent_start_pos[i])
            self.helper_grid.get(*agent_start_pos[i]).agent_above = True

    def _get_obs(self):
        """Get observation vectors of all agents in the environment.

        Returns
        -------
        agent_obs: list(ndarray)
            list of agent observations where the element at i^th list index is the observation vector for the agent with index i.
        """
        # initialize list of observation vector of each agent
        agent_obs = [
            np.zeros(
                (
                    self.obs_depth,
                    self.grid_size_without_walls + 1,
                    self.grid_size_without_walls + 1,
                ),
                dtype=np.float32,
            )
            for _ in range(self.num_agents)
        ]

        # update walls and tree states in agent observations
        for obj in self.helper_grid.grid:
            if obj is None:
                continue
            i, j = obj.pos
            for a in self.agents:
                # convert to agent centered coordinates
                nc = [i - a.pos[0], j - a.pos[1]]
                # wrap around to get agent centered toroidal coordinates
                if nc[0] < 0:
                    nc[0] += self.grid_size_without_walls + 1
                if nc[1] < 0:
                    nc[1] += self.grid_size_without_walls + 1
                # update agent's observation. # switch x and y coordinates because the y-coordinate specifies the row, while the x-coordinate specifies the column
                if obj.type == "tree":
                    agent_obs[a.index][obj.state, nc[1], nc[0]] = 1
                elif obj.type == "wall":
                    agent_obs[a.index][
                        len(STATE_IDX_TO_COLOR_WILDFIRE), nc[1], nc[0]
                    ] = 1

        # for each agent, update other agents' positions in agent observations
        for a in self.agents:
            for o in self.agents:
                if o.index != a.index:
                    idx = o.index - int(np.heaviside(o.index - a.index, 0))
                    # convert to agent centered coordinates
                    nc = [
                        o.pos[0] - a.pos[0],
                        o.pos[1] - a.pos[1],
                    ]
                    # wrap around to get agent centered toroidal coordinates
                    if nc[0] < 0:
                        nc[0] += self.grid_size_without_walls + 1
                    if nc[1] < 0:
                        nc[1] += self.grid_size_without_walls + 1
                    agent_obs[a.index][
                        len(STATE_IDX_TO_COLOR_WILDFIRE) + 1 + idx,
                        nc[1],
                        nc[0],
                    ] = 1

        # flatten, and append normalized time step at the end of, each agent observation
        for a in self.agents:
            agent_obs[a.index] = np.append(
                agent_obs[a.index].flatten(),
                np.array(self.step_count / self.max_steps, dtype=np.float32),
            )
        return agent_obs

    def get_state(self):
        """Get the state representation of the environment.

        Returns
        -------
        ndarray
            state representation of the environment
        """
        # initialize array to store state vector
        s = np.zeros(
            (
                self.obs_depth + 1,
                self.grid_size,
                self.grid_size,
            ),
            dtype=np.float32,
        )

        # update tree states and walls in state representation
        for o in self.helper_grid.grid:
            # switch x and y coordinates because the y-coordinate specifies the row, while the x-coordinate specifies the column
            if o.type == "tree":
                s[o.state, o.pos[1], o.pos[0]] = 1
            if o.type == "wall":
                s[len(STATE_IDX_TO_COLOR_WILDFIRE), o.pos[1], o.pos[0]] = 1

        # update agent positions in state representation
        for a in self.agents:
            s[
                len(STATE_IDX_TO_COLOR_WILDFIRE) + 1 + a.index,
                a.pos[1],
                a.pos[0],
            ] = 1

        # flatten, and append normalized time step at the end of, state representation
        s = np.append(
            s.flatten(),
            np.array(self.step_count / self.max_steps, dtype=np.float32),
        )
        return s

    def get_state_interpretation(self, state, print_interpretation=True):
        """Get human readable interpretation of the state of the environment

        Parameters
        ----------
        state : ndarray
            state representation of the environment
        print_interpretation : bool, optional
           whether to print the interpretation of the state, by default True. Human readable interpretation refers to printing the positions of trees on fire, agents, and the time step.

        Returns
        -------
        on_fire_trees : list[tuple[int,int]]
            list of tuples containing position coordinates (x,y) of trees on fire. Coordinates are in grid without wall coordinates
        time_step : float
            normalized time step of the episode at which time the state was recorded
        """
        time_step = state[-1]
        state = state[:-1].reshape(
            (
                self.obs_depth + 1,
                self.grid_size,
                self.grid_size,
            ),
        )
        if print_interpretation:
            print("-------------------------------------------------------------")
            print("State interpretation:")
        on_fire_trees = []
        for i in range(self.grid_size):
            for j in range(self.grid_size):
                if state[1, j, i] == 1:
                    if print_interpretation:
                        print(f"Tree at position {(i,j)} is on fire.")
                    on_fire_trees.append((i, j))
                for o in self.agents:
                    index = o.index
                    if state[len(STATE_IDX_TO_COLOR_WILDFIRE) + 1 + index, j, i] == 1:
                        if print_interpretation:
                            print(f"Agent {o.index} is at position {(i,j)}.")
        if print_interpretation:
            print(f"Time step: {time_step}")
            print("-------------------------------------------------------------")
        return on_fire_trees, time_step

    def construct_state(self, trees_on_fire, agent_pos, time_step: int):
        """Construct the state representation vector of the environment for given positions of trees on fire and agents

        Parameters
        ----------
        trees_on_fire : list
            list of tuples containing position coordinates (x,y) of trees on fire. Coordinates should be in grid without wall coordinates
        agent_pos : list
            list of tuples containing position coordinates (x,y) of agents, in order of agent index. Coordinates should be in grid without wall coordinates
        time_step : int
            normalized time step of the episode at which time the state was recorded

        Returns
        -------
        state : ndarray
            state representation of the environment
        """
        state = np.zeros(
            (
                self.obs_depth + 1,
                self.grid_size,
                self.grid_size,
            ),
            dtype=np.float32,
        )
        # update tree states and walls in state representation. There are no burnt trees in the state
        state[0, :, :] = 1
        for pos in trees_on_fire:
            state[1, pos[1], pos[0]] = 1
            state[0, pos[1], pos[0]] = 0
        for i in range(self.grid_size):
            for j in range(self.grid_size):
                if (
                    i == 0
                    or i == self.grid_size - 1
                    or j == 0
                    or j == self.grid_size - 1
                ):
                    state[len(STATE_IDX_TO_COLOR_WILDFIRE), j, i] = 1
                    state[0, j, i] = 0

        # update agent positions in state representation
        for i, pos in enumerate(agent_pos):
            state[
                len(STATE_IDX_TO_COLOR_WILDFIRE) + 1 + i,
                pos[1],
                pos[0],
            ] = 1

        # flatten, and append normalized time step at the end of, state representation
        state = np.append(
            state.flatten(),
            np.array(time_step / self.max_steps, dtype=np.float32),
        )
        return state

    def reset(self, seed=None, state=None):
        """Reset the state of the environment

        Parameters
        ----------
        seed : int, optional
            seed for random number generator, by default None
        state : ndarray, optional
            specifies the initial state of the environment upon reset, by default None.
            If none, initial state is chosen uniformly at random from initial state distribution.

        Returns
        -------
        obs : OrderedDict
            dictionary where each key is the agent index and the value is the observation vector for that agent.
        info : dict
            dictionary containing additional information about the environment.
            Here, it contains the number of burnt trees in the environment after reset
        """
        # reset environment attributes
        self.burnt_trees = 0
        self.trees_on_fire = 0
        self.unburnt_trees = []
        if self.log_selfish_region_metrics:
            self.selfish_region_trees_on_fire = np.zeros(len(self.selfish_xmin))
            self.selfish_region_burnt_trees = np.zeros(len(self.selfish_xmin))

        # reset the grid
        if state is not None:
            super().reset(seed=seed, state=state)
        else:
            super().reset(seed=seed)

        # get agent observations
        agent_obs = self._get_obs()
        obs = OrderedDict({f"{a.index}": agent_obs[a.index] for a in self.agents})

        # create info dictionary
        info = {"burnt trees": self.burnt_trees}
        return obs, info

    def move_agent(self, i, next_pos):
        """Move agent to a new position in the grid

        Parameters
        ----------
        i : int
            index of agent to be moved
        next_pos : tuple[int, int]
            coordinates of new position
        """
        # add agent to grid in new position
        self.grid.set(*next_pos, self.agents[i])

        # get tree in agent's old position from helper grid and add tree to grid
        tree = self.helper_grid.get(*self.agents[i].pos)
        tree.agent_above = False
        self.grid.set(*self.agents[i].pos, tree)

        # update attributes
        next_tree = self.helper_grid.get(*next_pos)
        next_tree.agent_above = True
        self.agents[i].pos = next_pos

    def neighbors_on_fire(self, tree_pos) -> int:
        """Get the number of neighboring trees on fire for a given tree.
           Neighbors are adjacent trees in the cardinal directions. A tree can have at most 4 neighbors

        Parameters
        ----------
        tree_pos : tuple[int, int]
            position coordinates of the tree whose neighbors are to be checked

        Returns
        -------
        num : int
            the number of neighboring trees on fire
        """
        num = 0
        relative_pos = [
            np.array([1, 0]),
            np.array([-1, 0]),
            np.array([0, 1]),
            np.array([0, -1]),
        ]
        for r in relative_pos:
            neighbor_pos = tree_pos + r
            if neighbor_pos[0] >= 0 and neighbor_pos[0] < self.helper_grid.width:
                if neighbor_pos[1] >= 0 and neighbor_pos[1] < self.helper_grid.height:
                    o = self.helper_grid.get(*neighbor_pos)
                    if o is not None and o.type == "tree":
                        if o.state == 1:
                            num += 1
        return num

    def in_selfish_region(self, i: int, j: int, region_index: int) -> bool:
        """Check if given tree is in region of selfish interest with given index

        Parameters
        ----------
        i : int
            x-coordinate of tree position
        j : int
            y-coordinate of tree position
        region_index : int
            index of selfish region. Same as index of corresponding selfish agent (or group of selfish agents).

        Returns
        -------
        bool
            True, if tree is in the region of selfish interest. Otherwise, False
        """
        return (
            i >= self.selfish_xmin[region_index]
            and i <= self.selfish_xmax[region_index]
            and j >= self.selfish_ymin[region_index]
            and j <= self.selfish_ymax[region_index]
        )

    def step(self, actions):
        """Take a step in the environment. Wildfire dynamics are propagated by one time step, and agents move according to their actions.

        Parameters
        ----------
        actions : dict
            dictionary containing actions for each agent. The key is the agent index and the value is the action.

        Returns
        -------
        next_obs : OrderedDict
            dictionary where each key is the agent index and the value is the new observation vector (ndarray) for that agent after the environment step.
        rewards : dict
            dictionary where each key is the agent index and the value is the reward given to that agent after the environment step.
        terminated : bool
            True, if the episode is done, otherwise, False. Episode is done if maximum number of steps is reached or there are zero trees on fire.
        info : dict
            dictionary where each key is the agent index and the value is an info dictionary containing additional information about the environment. Here, each agent's info dictionary contains the same information, viz., the number of burnt trees.
        """
        self.step_count += 1
        actions = list(actions.values())
        terminated = False
        truncated = False

        # Move agents sequentially, in random order
        order = np.random.permutation(len(actions))
        for i in order:
            if actions[i] == self.actions.STILL:
                continue
            if actions[i] == self.actions.NORTH:
                next_pos = self.agents[i].north_pos()
                next_cell = self.grid.get(*next_pos)
                if next_cell is None or next_cell.can_overlap():
                    self.move_agent(i, next_pos)
            elif actions[i] == self.actions.SOUTH:
                next_pos = self.agents[i].south_pos()
                next_cell = self.grid.get(*next_pos)
                if next_cell is None or next_cell.can_overlap():
                    self.move_agent(i, next_pos)
            elif actions[i] == self.actions.EAST:
                next_pos = self.agents[i].east_pos()
                next_cell = self.grid.get(*next_pos)
                if next_cell is None or next_cell.can_overlap():
                    self.move_agent(i, next_pos)
            elif actions[i] == self.actions.WEST:
                next_pos = self.agents[i].west_pos()
                next_cell = self.grid.get(*next_pos)
                if next_cell is None or next_cell.can_overlap():
                    self.move_agent(i, next_pos)

        # propagate wildfire dynamics by one time step
        # initialize lists to store trees transitioning to on fire and burnt state in the current time step
        trees_to_fire_state = []
        if self.log_selfish_region_metrics:
            num_trees_to_fire_state_sr = {
                f"{i}": 0 for i, _ in enumerate(self.selfish_xmin)
            }
        trees_to_burnt_state = []
        # loop over all unburnt trees. Burnt trees remain burnt
        for c in self.unburnt_trees:
            if c.state == 0:
                pos = np.array(c.pos)
                # transition from healthy to on fire with probability 1 - (1 - alpha)^n
                if np.random.rand() < 1 - (1 - self.alpha) ** self.neighbors_on_fire(
                    pos
                ):
                    # update relevant attributes and lists
                    trees_to_fire_state.append(c)
                    self.trees_on_fire += 1
                    if self.log_selfish_region_metrics:
                        if c.region != "common":
                            self.selfish_region_trees_on_fire[int(c.region)] += 1
                            num_trees_to_fire_state_sr[c.region] += 1
            if c.state == 1:
                # transition from on fire to burnt with probability 1 - beta + delta_beta * agent_above
                if np.random.rand() < 1 - self.beta + c.agent_above * self.delta_beta:
                    # update relevant attributes and list
                    trees_to_burnt_state.append(c)
                    self.burnt_trees += 1
                    self.trees_on_fire -= 1
                    if self.log_selfish_region_metrics:
                        if c.region != "common":
                            self.selfish_region_burnt_trees[int(c.region)] += 1
                            self.selfish_region_trees_on_fire[int(c.region)] -= 1

        # update tree objects in helper grid and grid. This update is done after the loop to avoid affecting
        # the transition probabilities of trees later in the loop due to trees that have already transitioned earlier in the loop.
        for c in trees_to_fire_state:
            c.state = 1
            c.color = STATE_IDX_TO_COLOR_WILDFIRE[c.state]
            # update tree in grid
            o = self.grid.get(c.pos[0], c.pos[1])
            if o.type == "tree":
                o.state = 1
                o.color = STATE_IDX_TO_COLOR_WILDFIRE[o.state]
        for c in trees_to_burnt_state:
            self.unburnt_trees.remove(c)
            c.state = 2
            c.color = STATE_IDX_TO_COLOR_WILDFIRE[c.state]
            # update tree in grid
            o = self.grid.get(c.pos[0], c.pos[1])
            if o.type == "tree":
                o.state = 2
                o.color = STATE_IDX_TO_COLOR_WILDFIRE[o.state]

        # check if episode is done
        if self.trees_on_fire == 0:
            terminated = True
            rewards = {f"{a.index}": 0 for a in self.agents}
        elif self.step_count >= self.max_steps:
            truncated = True
            rewards = {f"{a.index}": 0 for a in self.agents}
        else:
            # compute agent rewards
            agent_rewards = np.zeros(self.num_agents)
            if self.cooperative_reward:
                agent_rewards -= 0.5 * len(trees_to_fire_state)
            else:
                if self.agent_groups:
                    for a in self.agents:
                        agent_rewards[a.index] -= 0.5 * num_trees_to_fire_state_sr[
                            f"{self.idx_to_group[a.index]}"
                        ] + 0.1 * (
                            len(trees_to_fire_state)
                            - num_trees_to_fire_state_sr[
                                f"{self.idx_to_group[a.index]}"
                            ]
                        )
                else:
                    for a in self.agents:
                        agent_rewards[a.index] -= 0.5 * num_trees_to_fire_state_sr[
                            f"{a.index}"
                        ] + 0.1 * (
                            len(trees_to_fire_state)
                            - num_trees_to_fire_state_sr[f"{a.index}"]
                        )
            # agent rewards dictionary
            rewards = {f"{a.index}": agent_rewards[a.index] for a in self.agents}

        # get agent observations after the environment step
        agent_obs = self._get_obs()
        next_obs = OrderedDict({f"{a.index}": agent_obs[a.index] for a in self.agents})

        # info dictionary
        info = {"burnt trees": self.burnt_trees}
        infos = {f"{a.index}": info for a in self.agents}

        return next_obs, rewards, terminated or truncated, infos

    def render(self, dummy_arg, close=False, highlight=False, tile_size=TILE_PIXELS):
        """Render the whole-grid human view

        Parameters
        ----------
        dummy_arg : Any
            temporary dummy argument to fix a bug where the first argument to render function call is 'human' even though it is not passed in code itself.
        close : bool, optional
            close the rendering window, by default False. Only applicable if render_mode is "human"
        highlight : bool, optional
            highlight the cells visible to the agent, by default False
        tile_size : int, optional
            size of each tile in pixels, by default TILE_PIXELS (defined in constants.py)

        Returns
        -------
        img : ndarray
            image of the grid
        """
        if close:
            if self.window:
                self.window.close()
            return

        if self.render_mode == "human" and not self.window:
            self.window = Window("gym_multigrid")
            self.window.show(block=False)

        if highlight:
            # Compute which cells are visible to the agent
            _, vis_masks = self.gen_obs_grid()

            highlight_masks = {
                (i, j): [] for i in range(self.width) for j in range(self.height)
            }

            for i, a in enumerate(self.agents):
                # Compute the world coordinates of the bottom-left corner
                # of the agent's view area
                f_vec = a.dir_vec
                r_vec = a.right_vec
                top_left = (
                    a.pos + f_vec * (a.view_size - 1) - r_vec * (a.view_size // 2)
                )

                # Mask of which cells to highlight

                # For each cell in the visibility mask
                for vis_j in range(0, a.view_size):
                    for vis_i in range(0, a.view_size):
                        # If this cell is not visible, don't highlight it
                        if not vis_masks[i][vis_i, vis_j]:
                            continue

                        # Compute the world coordinates of this cell
                        abs_i, abs_j = top_left - (f_vec * vis_j) + (r_vec * vis_i)

                        if abs_i < 0 or abs_i >= self.width:
                            continue
                        if abs_j < 0 or abs_j >= self.height:
                            continue

                        # Mark this cell to be highlighted
                        highlight_masks[abs_i, abs_j].append(i)

        # Render the grid
        if self.render_selfish_region_boundaries:
            # include selfish region boundaries in the render
            colors = [COLORS[color] for color in self.agent_colors]
            img = self.grid.render(
                tile_size,
                highlight_masks=highlight_masks if highlight else None,
                uncached_object_types=self.uncahed_object_types,
                x_min=self.selfish_xmin,
                y_min=self.selfish_ymin,
                x_max=self.selfish_xmax,
                y_max=self.selfish_ymax,
                colors=colors,
            )
        else:
            img = self.grid.render(
                tile_size,
                highlight_masks=highlight_masks if highlight else None,
                uncached_object_types=self.uncahed_object_types,
            )

        # Re-render the tiles containing agents to include trees below agent
        if self.render_selfish_region_boundaries:
            # include selfish region boundaries in the render
            for a in self.agents:
                img = render_agent_tiles(
                    img,
                    a,
                    self.helper_grid,
                    self.world,
                    x_min=self.selfish_xmin,
                    y_min=self.selfish_ymin,
                    x_max=self.selfish_xmax,
                    y_max=self.selfish_ymax,
                    colors=colors,
                )
        else:
            for a in self.agents:
                img = render_agent_tiles(img, a, self.helper_grid, self.world)

        if self.render_mode == "human":
            self.window.show_img(img)

        return img



import enum
import math
from typing import TypeVar

import numpy as np
from wildfire_environment.core.grid import Grid
from wildfire_environment.core.world import WorldT
from wildfire_environment.policy.base import AgentPolicyT
from wildfire_environment.utils.rendering import (
    point_in_triangle,
    rotate_fn,
    fill_coords,
)
from wildfire_environment.core.object import WorldObj
from wildfire_environment.core.constants import DIR_TO_VEC

ActionsT = TypeVar("ActionsT", bound=enum.IntEnum)


class DefaultActions(enum.IntEnum):
    """Set of DefaultActions

    Parameters
    ----------
    enum : IntEnum
        Base class for creating enumerated constants that are also subclasses of int.
    """

    STILL = 0
    LEFT = 1
    RIGHT = 2
    FORWARD = 3
    PICKUP = 4
    DROP = 5
    TOGGLE = 6
    DONE = 7


class WildfireActions(enum.IntEnum):
    """Set of actions available for the agents in Wildfire environment

    Parameters
    ----------
    enum : IntEnum
        Base class for creating enumerated constants that are also subclasses of int.
    """

    STILL = 0
    NORTH = 1
    EAST = 2
    SOUTH = 3
    WEST = 4


AgentT = TypeVar("AgentT", bound="Agent")


class Agent(WorldObj):
    """Defines the class for an agent in the environment"""

    def __init__(
        self,
        world: WorldT,
        index: int = 0,
        view_size: int = 7,
        actions=DefaultActions,
        dir_to_vec=DIR_TO_VEC,
        color=None,
        bg_color=None,
        type: str = "agent",
    ):
        """Initialize the agent object

        Parameters
        ----------
        world : WorldT
            the world within which grid is situated
        index : int, optional
            a number useful to identify the instantiated agent, by default 0
        view_size : int, optional
            the size of agent view, if partial observability holds, by default 7
        actions : Type[ActionsT], optional
            set of actions available to the agent, by default DefaultActions
        dir_to_vec : list[NDArray], optional
            map of agent direction indices to vectors, by default DIR_TO_VEC
        color : str | None, optional
            color of the agent, by default None
        bg_color : str | None, optional
            background color of the tile containing agent, by default None
        type : str, optional
            type of the world object, by default "agent"
        """
        if color is None:
            color = world.IDX_TO_COLOR[index]
        else:
            pass

        super().__init__(world, type, color, bg_color)
        self.pos = None
        self.dir = None
        self.init_dir = None
        self.index = index
        self.view_size = view_size
        self.carrying = None
        self.terminated = False
        self.started = True
        self.paused = False
        self.actions = actions
        self.world = world
        self.dir_to_vec = dir_to_vec

    def render(self, img):
        """Render the agent at its current position

        Parameters
        ----------
        img : NDArray
            the image to render the agent in
        """
        c = self.world.COLORS[self.color]
        tri_fn = point_in_triangle(
            (0.12, 0.19),
            (0.87, 0.50),
            (0.12, 0.81),
        )
        # Rotate the agent based on its direction
        assert self.dir is not None
        tri_fn = rotate_fn(tri_fn, cx=0.5, cy=0.5, theta=0.5 * math.pi * self.dir)
        fill_coords(
            img, tri_fn, c, self.world.COLORS[self.bg_color] if self.bg_color else None
        )

    def encode(self, current_agent=False):
        """Encode a description of this object as a 3-tuple of integers

        Parameters
        ----------
        current_agent : bool, optional
            whether the agent is the current agent, by default False
        """
        if self.world.encode_dim == 3:
            return (
                self.world.OBJECT_TO_IDX[self.type],
                self.world.COLOR_TO_IDX[self.color],
                self.dir,
            )
        elif self.carrying:
            if current_agent:
                return (
                    self.world.OBJECT_TO_IDX[self.type],
                    self.world.COLOR_TO_IDX[self.color],
                    self.world.OBJECT_TO_IDX[self.carrying.type],
                    self.world.COLOR_TO_IDX[self.carrying.color],
                    self.dir,
                    1,
                )
            else:
                return (
                    self.world.OBJECT_TO_IDX[self.type],
                    self.world.COLOR_TO_IDX[self.color],
                    self.world.OBJECT_TO_IDX[self.carrying.type],
                    self.world.COLOR_TO_IDX[self.carrying.color],
                    self.dir,
                    0,
                )

        else:
            if current_agent:
                return (
                    self.world.OBJECT_TO_IDX[self.type],
                    self.world.COLOR_TO_IDX[self.color],
                    0,
                    0,
                    self.dir,
                    1,
                )
            else:
                return (
                    self.world.OBJECT_TO_IDX[self.type],
                    self.world.COLOR_TO_IDX[self.color],
                    0,
                    0,
                    self.dir,
                    0,
                )

    def move(
        self,
        next_pos,
        grid: Grid,
        init_grid=None,
        dummy_move: bool = False,
        bg_color=None,
    ):
        """Move the agent to a new position

        Parameters
        ----------
        next_pos : Position
            the position to move the agent to
        grid : Grid
            the grid to move the agent in
        init_grid : Grid | None, optional
            the initial grid before agent is moved, by default None
        dummy_move : bool, optional
            whether the move is a dummy move, by default False
        bg_color : str | None, optional
            the background color of the tile containing agent, by default None
        """
        if self.pos is not None:
            direction = np.array(next_pos) - np.array(self.pos)

            for i, vec in enumerate(self.dir_to_vec):
                if np.array_equal(vec, direction):
                    self.dir = i
                    break

            if init_grid is not None:
                grid.set(*self.pos, init_grid.get(*self.pos))
        else:
            pass

        if dummy_move:
            pass
        else:
            self.pos = next_pos

        assert self.pos is not None
        grid.set(*self.pos, self)

        if bg_color is not None:
            self.bg_color = bg_color
        else:
            pass

    @property
    def dir_vec(self):
        """
        Get the direction vector for the agent, pointing in the direction
        of forward movement.

        Returns
        -------
        NDArray
            the direction vector for forward movement in the current orientation of the agent
        """

        assert self.dir is not None
        assert self.dir >= 0 and self.dir < len(self.dir_to_vec)
        return self.dir_to_vec[self.dir]

    @property
    def right_vec(self):
        """
        Get the vector pointing to the right of the agent.

        Returns
        -------
        NDArray
            the vector pointing to the right of the agent
        """

        dx, dy = self.dir_vec
        return np.array((-dy, dx))

    @property
    def front_pos(self):
        """
        Get the position of the cell that is right in front of the agent

        Returns
        -------
        NDArray
            the position of the cell that is right in front of the agent
        """

        return self.pos + self.dir_vec

    def west_pos(self):
        """
        Get the position of the cell to the left of the agent

        Returns
        -------
        NDArray
            the position of the cell to the left of the agent
        """
        if self.pos is None:
            raise ValueError("Agent position is not set")
        else:
            return self.pos + np.array([-1, 0])

    def east_pos(self):
        """
        Get the position of the cell to the right of the agent

        Returns
        -------
        NDArray
            the position of the cell to the right of the agent
        """
        if self.pos is None:
            raise ValueError("Agent position is not set")
        else:
            return self.pos + np.array([1, 0])

    def north_pos(self):
        """
        Get the position of the cell above the agent

        Returns
        -------
        NDArray
            the position of the cell above the agent
        """
        if self.pos is None:
            raise ValueError("Agent position is not set")
        else:
            return self.pos + np.array([0, -1])

    def south_pos(self):
        """
        Get the position of the cell below the agent

        Returns
        -------
        NDArray
            the position of the cell below the agent
        """
        if self.pos is None:
            raise ValueError("Agent position is not set")
        else:
            return self.pos + np.array([0, 1])

    def get_view_coords(self, i, j):
        """
        Translate and rotate absolute grid coordinates (i, j) into the
        agent's partially observable view (sub-grid). Note that the resulting
        coordinates may be negative or outside of the agent's view size.

        Parameters
        ----------
        i : int
            x-coordinate in the grid
        j : int
            y-coordinate in the grid

        Returns
        -------
        tuple
            the coordinates of the grid in the agent's view
        """

        assert self.pos is not None
        ax, ay = self.pos
        dx, dy = self.dir_vec
        rx, ry = self.right_vec

        # Compute the absolute coordinates of the top-left view corner
        sz = self.view_size
        hs = self.view_size // 2
        tx = ax + (dx * (sz - 1)) - (rx * hs)
        ty = ay + (dy * (sz - 1)) - (ry * hs)

        lx = i - tx
        ly = j - ty

        # Project the coordinates of the object relative to the top-left
        # corner onto the agent's own coordinate system
        vx = rx * lx + ry * ly
        vy = -(dx * lx + dy * ly)

        return vx, vy

    def get_view_exts(self):
        """
        Get the extents of the square set of tiles visible to the agent
        Note: the bottom extent indices are not included in the set

        Returns
        -------
        tuple
            the extents of the square set of tiles visible to the agent
        """

        assert self.pos is not None

        # Facing right
        if self.dir == 0:
            topX = self.pos[0]
            topY = self.pos[1] - self.view_size // 2
        # Facing down
        elif self.dir == 1:
            topX = self.pos[0] - self.view_size // 2
            topY = self.pos[1]
        # Facing left
        elif self.dir == 2:
            topX = self.pos[0] - self.view_size + 1
            topY = self.pos[1] - self.view_size // 2
        # Facing up
        elif self.dir == 3:
            topX = self.pos[0] - self.view_size // 2
            topY = self.pos[1] - self.view_size + 1
        else:
            assert False, "invalid agent direction"

        botX = topX + self.view_size
        botY = topY + self.view_size

        return (topX, topY, botX, botY)

    def relative_coords(self, x, y):
        """
        Check if a grid position belongs to the agent's field of view, and returns the corresponding coordinates

        Parameters
        ----------
        x : int
            x-coordinate in the grid
        y : int
            y-coordinate in the grid

        Returns
        -------
        tuple | None
            the coordinates of the grid position in the agent's view or None if the position is not visible
        """

        vx, vy = self.get_view_coords(x, y)

        if vx < 0 or vy < 0 or vx >= self.view_size or vy >= self.view_size:
            return None

        return vx, vy

    def in_view(self, x, y):
        """
        check if a grid position is visible to the agent

        Parameters
        ----------
        x : int
            x-coordinate in the grid
        y : int
            y-coordinate in the grid

        Returns
        -------
        bool
            whether the grid position is visible to the agent
        """

        return self.relative_coords(x, y) is not None


class PolicyAgent(Agent):
    """
    Agent with a policy that determines its actions
    """

    def __init__(
        self,
        policy: AgentPolicyT,
        world: WorldT,
        index: int = 0,
        view_size: int = 7,
        actions=DefaultActions,
        dir_to_vec=DIR_TO_VEC,
        color=None,
        bg_color=None,
        type: str = "agent",
    ):
        """Initialize the PolicyAgent object

        Parameters
        ----------
        policy : AgentPolicyT
            the policy that determines the agent's actions
        world : WorldT
            the world within which grid is situated
        index : int, optional
            a number useful to identify the instantiated agent, by default 0
        view_size : int, optional
            the size of agent view, if partial observability holds, by default 7
        actions : type[ActionsT], optional
            set of actions available to the agent, by default DefaultActions
        dir_to_vec : list[NDArray], optional
            map of agent direction indices to vectors, by default DIR_TO_VEC
        color : str | None, optional
            color of the agent, by default None
        bg_color : str | None, optional
            background color of the tile containing agent, by default None
        type : str, optional
            type of the world object, by default "agent"
        """
        super().__init__(
            world, index, view_size, actions, dir_to_vec, color, bg_color, type
        )
        self.policy: AgentPolicyT = policy



import numpy as np

# Size in pixels of a tile in the full-scale human view
TILE_PIXELS = 32

# Map of color names to RGB values
COLORS = {
    "red": np.array([228, 3, 3]),
    "orange": np.array([255, 140, 0]),
    "yellow": np.array([255, 237, 0]),
    "green": np.array([0, 128, 38]),
    "blue": np.array([0, 77, 255]),
    "purple": np.array([117, 7, 135]),
    "brown": np.array([120, 79, 23]),
    "grey": np.array([100, 100, 100]),
    "light_red": np.array([234, 153, 153]),
    "light_blue": np.array([90, 170, 223]),
    "black": np.array([0, 0, 0]),
    "white": np.array([255, 255, 255]),
    "medium_red": np.array([231, 80, 80]),
}

COLOR_NAMES = sorted(list(COLORS.keys()))

# Used to map colors to integers
COLOR_TO_IDX = {key: i for i, key in enumerate(COLORS.keys())}

# Map of state names to integers
STATE_TO_IDX = {
    "open": 0,
    "closed": 1,
    "locked": 2,
}

# Map of state names to integers
STATE_TO_IDX_WILDFIRE = {
    "healthy": 0,
    "on fire": 1,
    "burnt": 2,
}

# Map of state idx to color
STATE_IDX_TO_COLOR_WILDFIRE = {
    0: "green",
    1: "orange",
    2: "brown",
}

# Map of agent direction indices to vectors
DIR_TO_VEC = [
    # Pointing right (positive X)
    np.array((1, 0)),
    # Down (positive Y)
    np.array((0, 1)),
    # Pointing left (negative X)
    np.array((-1, 0)),
    # Up (negative Y)
    np.array((0, -1)),
]

# Map of object types to short string
OBJECT_TO_STR = {
    "wall": "x",
    "floor": "F",
    "door": "D",
    "key": "K",
    "ball": "o",
    "box": "B",
    "goal": "G",
    "lava": "V",
    "agent": "a",
    "tree": "T",
}

# Short string for opened door
OPENED_DOOR_IDS = "_"

# Map agent's direction to short string
AGENT_DIR_TO_STR = {0: ">", 1: "V", 2: "<", 3: "^"}



from gym.envs.registration import register


# Wildfire environment
# ----------------------------------------
register(
    id="wildfire-v0",
    entry_point="wildfire_environment.envs:WildfireEnv",
)



from typing import TypeVar
from dataclasses import dataclass, field

from wildfire_environment.core.constants import COLORS

WorldT = TypeVar("WorldT", bound="World")


@dataclass
class World:
    """This class defines the world within which grid is situated."""

    encode_dim: int
    normalize_obs: int
    OBJECT_TO_IDX: dict  # Map of object type to integers
    COLORS: dict  # Map of color names to RGB values
    COLOR_TO_IDX: dict = field(init=False)
    IDX_TO_COLOR: dict = field(init=False)
    IDX_TO_OBJECT: dict = field(init=False)

    def __post_init__(self):
        self.COLOR_TO_IDX = dict(
            zip(self.COLORS.keys(), range(len(self.COLORS.keys())))
        )
        self.IDX_TO_COLOR = dict(
            zip(self.COLOR_TO_IDX.values(), self.COLOR_TO_IDX.keys())
        )
        self.IDX_TO_OBJECT = dict(
            zip(self.OBJECT_TO_IDX.values(), self.OBJECT_TO_IDX.keys())
        )


DefaultWorld = World(
    encode_dim=6,
    normalize_obs=1,
    COLORS=COLORS,
    OBJECT_TO_IDX={
        "unseen": 0,
        "empty": 1,
        "wall": 2,
        "floor": 3,
        "door": 4,
        "key": 5,
        "ball": 6,
        "box": 7,
        "goal": 8,
        "lava": 9,
        "agent": 10,
        "objgoal": 11,
        "switch": 12,
    },
)

CollectWorld = World(
    encode_dim=3,
    normalize_obs=1,
    COLORS=COLORS,
    OBJECT_TO_IDX={
        "empty": 0,
        "wall": 1,
        "ball": 2,
        "agent": 3,
    },
)

WildfireWorld = World(
    encode_dim=3,
    normalize_obs=1,
    COLORS=COLORS,
    OBJECT_TO_IDX={
        "empty": 0,
        "tree": 1,
        "agent": 2,
        "wall": 3,
    },
)



from typing import TypeVar
import numpy as np

from wildfire_environment.core.world import WorldT
from wildfire_environment.utils.rendering import *
from wildfire_environment.core.constants import STATE_IDX_TO_COLOR_WILDFIRE


WorldObjT = TypeVar("WorldObjT", bound="WorldObj")


class WorldObj:
    """
    Base class for grid world objects
    """

    def __init__(
        self,
        world: WorldT,
        type: str = "base",
        color: str = "grey",
        bg_color=None,
    ):
        """Create a WorldObj object

        Parameters
        ----------
        world : WorldT
            the world in which the object exists
        type : str, optional
            type of the object, by default "base"
        color : str, optional
            color of the object, by default "grey"
        bg_color : str | None, optional
            background color of the tile containing object, by default None
        """
        assert type in world.OBJECT_TO_IDX, type
        assert color in world.COLOR_TO_IDX, color
        self.type: str = type
        self.color: str = color
        self.contains = None
        self.world = world
        self.bg_color = bg_color

        # Initial position of the object
        self.init_pos = None

        # Current position of the object
        self.pos = None

    def can_overlap(self) -> bool:
        """Can the agent overlap with this?"""
        return False

    def can_pickup(self) -> bool:
        """Can the agent pick this up?"""
        return False

    def can_contain(self) -> bool:
        """Can this contain another object?"""
        return False

    def see_behind(self) -> bool:
        """Can the agent see behind this object?"""
        return True

    def toggle(self, env, pos):
        """Method to trigger/toggle an action this object performs"""
        return False

    def encode(self, current_agent: bool = False):
        """Encode the a description of this object as a 3-tuple of integers"""
        if self.world.encode_dim == 3:
            return (
                self.world.OBJECT_TO_IDX[self.type],
                self.world.COLOR_TO_IDX[self.color],
                0,
            )
        else:
            return (
                self.world.OBJECT_TO_IDX[self.type],
                self.world.COLOR_TO_IDX[self.color],
                0,
                0,
                0,
                0,
            )

    @staticmethod
    def decode(type_idx: int, color_idx: int, state: int):
        assert False, "not implemented"

    def render(self, r) -> None:
        """Draw this object with the given renderer"""
        raise NotImplementedError


class ObjectGoal(WorldObj):
    def __init__(
        self,
        world: WorldT,
        index: int,
        target_type: str = "ball",
        reward: float = 1,
        color=None,
    ):
        if color is None:
            super().__init__(world, "objgoal", world.IDX_TO_COLOR[index])
        else:
            super().__init__(world, "objgoal", world.IDX_TO_COLOR[color])
        self.target_type: str = target_type
        self.index: int = index
        self.reward: float = reward

    def can_overlap(self):
        return False

    def render(self, img):
        fill_coords(img, point_in_rect(0, 1, 0, 1), self.world.COLORS[self.color])


class Goal(WorldObj):
    def __init__(self, world: WorldT, index: int, reward=1, color=None):
        if color is None:
            super().__init__(world, "goal", world.IDX_TO_COLOR[index])
        else:
            super().__init__(world, "goal", world.IDX_TO_COLOR[color])
        self.index = index
        self.reward = reward

    def can_overlap(self):
        return True

    def render(self, img):
        fill_coords(img, point_in_rect(0, 1, 0, 1), self.world.COLORS[self.color])


class Switch(WorldObj):
    def __init__(self, world: WorldT):
        super().__init__(world, "switch", world.IDX_TO_COLOR[0])

    def can_overlap(self):
        return True

    def render(self, img):
        fill_coords(img, point_in_rect(0, 1, 0, 1), self.world.COLORS[self.color])


class Floor(WorldObj):
    """
    Colored floor tile the agent can walk over
    """

    def __init__(self, world: WorldT, color: str = "blue", type: str = "floor"):
        super().__init__(world, type, color)

    def can_overlap(self) -> bool:
        return True

    def render(self, img):
        fill_coords(img, point_in_rect(0, 1, 0, 1), self.world.COLORS[self.color])


class Lava(WorldObj):
    def __init__(self, world: WorldT):
        super().__init__(world, "lava", "red")

    def can_overlap(self):
        return True

    def render(self, img):
        c = (255, 128, 0)

        # Background color
        fill_coords(img, point_in_rect(0, 1, 0, 1), c)

        # Little waves
        for i in range(3):
            ylo = 0.3 + 0.2 * i
            yhi = 0.4 + 0.2 * i
            fill_coords(img, point_in_line(0.1, ylo, 0.3, yhi, r=0.03), (0, 0, 0))
            fill_coords(img, point_in_line(0.3, yhi, 0.5, ylo, r=0.03), (0, 0, 0))
            fill_coords(img, point_in_line(0.5, ylo, 0.7, yhi, r=0.03), (0, 0, 0))
            fill_coords(img, point_in_line(0.7, yhi, 0.9, ylo, r=0.03), (0, 0, 0))


class Wall(WorldObj):
    def __init__(self, world: WorldT, color: str = "grey"):
        super().__init__(world, "wall", color)

    def see_behind(self):
        return False

    def render(self, img):
        fill_coords(img, point_in_rect(0, 1, 0, 1), self.world.COLORS[self.color])


class Obstacle(WorldObj):
    def __init__(
        self,
        world: WorldT,
        penalty: float = 0,
        can_see_through: bool = True,
        color: str = "grey",
    ):
        super().__init__(world, "obstacle", color)
        self.penalty = penalty
        self.can_see_through = can_see_through

    def see_behind(self):
        return self.can_see_through

    def can_overlap(self):
        return True if self.penalty != 0 else False

    def render(self, img):
        fill_coords(img, point_in_rect(0, 1, 0, 1), self.world.COLORS[self.color])


class Door(WorldObj):
    def __init__(
        self,
        world: WorldT,
        color: str,
        is_open: bool = False,
        is_locked: bool = False,
    ):
        super().__init__(world, "door", color)
        self.is_open: bool = is_open
        self.is_locked: bool = is_locked

    def can_overlap(self):
        """The agent can only walk over this cell when the door is open"""
        return self.is_open

    def see_behind(self):
        return self.is_open

    def toggle(self, env, pos):
        # If the player has the right key to open the door
        if self.is_locked:
            if isinstance(env.carrying, Key) and env.carrying.color == self.color:
                self.is_locked = False
                self.is_open = True
                return True
            return False

        self.is_open = not self.is_open
        return True

    def encode(self, current_agent: bool = False):
        """Encode the a description of this object as a 3-tuple of integers"""

        # State, 0: open, 1: closed, 2: locked
        state: int
        if self.is_open:
            state = 0
        elif self.is_locked:
            state = 2
        elif not self.is_open:
            state = 1
        else:
            raise ValueError("Invalid door state")

        return (
            self.world.OBJECT_TO_IDX[self.type],
            self.world.COLOR_TO_IDX[self.color],
            state,
            0,
            0,
            0,
        )

    def render(self, img):
        c = self.world.COLORS[self.color]

        if self.is_open:
            fill_coords(img, point_in_rect(0.88, 1.00, 0.00, 1.00), c)
            fill_coords(img, point_in_rect(0.92, 0.96, 0.04, 0.96), (0, 0, 0))
            return

        # Door frame and door
        if self.is_locked:
            fill_coords(img, point_in_rect(0.00, 1.00, 0.00, 1.00), c)
            fill_coords(img, point_in_rect(0.06, 0.94, 0.06, 0.94), 0.45 * np.array(c))

            # Draw key slot
            fill_coords(img, point_in_rect(0.52, 0.75, 0.50, 0.56), c)
        else:
            fill_coords(img, point_in_rect(0.00, 1.00, 0.00, 1.00), c)
            fill_coords(img, point_in_rect(0.04, 0.96, 0.04, 0.96), (0, 0, 0))
            fill_coords(img, point_in_rect(0.08, 0.92, 0.08, 0.92), c)
            fill_coords(img, point_in_rect(0.12, 0.88, 0.12, 0.88), (0, 0, 0))

            # Draw door handle
            fill_coords(img, point_in_circle(cx=0.75, cy=0.50, r=0.08), c)


class Key(WorldObj):
    def __init__(self, world: WorldT, color: str = "blue"):
        super(Key, self).__init__(world, "key", color)

    def can_pickup(self):
        return True

    def render(self, img):
        c = self.world.COLORS[self.color]

        # Vertical quad
        fill_coords(img, point_in_rect(0.50, 0.63, 0.31, 0.88), c)

        # Teeth
        fill_coords(img, point_in_rect(0.38, 0.50, 0.59, 0.66), c)
        fill_coords(img, point_in_rect(0.38, 0.50, 0.81, 0.88), c)

        # Ring
        fill_coords(img, point_in_circle(cx=0.56, cy=0.28, r=0.190), c)
        fill_coords(img, point_in_circle(cx=0.56, cy=0.28, r=0.064), (0, 0, 0))


class Ball(WorldObj):
    def __init__(self, world, index=0, reward=2):
        super().__init__(world, "ball", world.IDX_TO_COLOR[index])
        self.index = index
        self.reward = reward

    def can_pickup(self):
        return True

    def can_overlap(self):
        return True

    def render(self, img):
        fill_coords(img, point_in_circle(0.5, 0.5, 0.31), self.world.COLORS[self.color])


class Box(WorldObj):
    def __init__(self, world: WorldT, color: str, contains=None):
        super(Box, self).__init__(world, "box", color)
        self.contains = contains

    def can_pickup(self):
        return True

    def render(self, img):
        c = self.world.COLORS[self.color]

        # Outline
        fill_coords(img, point_in_rect(0.12, 0.88, 0.12, 0.88), c)
        fill_coords(img, point_in_rect(0.18, 0.82, 0.18, 0.82), (0, 0, 0))

        # Horizontal slit
        fill_coords(img, point_in_rect(0.16, 0.84, 0.47, 0.53), c)

    def toggle(self, env, pos):
        # Replace the box by its contents
        env.grid.set(*pos, self.contains)
        return True


class Flag(WorldObj):
    def __init__(
        self,
        world: WorldT,
        index: int,
        type: str = "flag",
        color: str = "blue",
        bg_color: str = "light_blue",
    ):
        super().__init__(world, type, color, bg_color)
        self.index: int = index

    def can_pickup(self):
        return True

    def can_overlap(self):
        return True

    def render(self, img):
        fill_coords(
            img,
            point_in_circle(0.5, 0.5, 0.31),
            self.world.COLORS[self.color],
            self.world.COLORS[self.bg_color] if self.bg_color else None,
        )


class Tree(WorldObj):
    def __init__(
        self,
        world: WorldT,
        tree_state_idx: int = 0,
        region: str = "common",
    ):
        super().__init__(world, "tree", STATE_IDX_TO_COLOR_WILDFIRE[tree_state_idx])
        self.state = tree_state_idx
        self.agent_above = False
        self.region = region

    def can_overlap(self):
        return True

    def encode(self, current_agent: bool = False):
        return (
            self.world.OBJECT_TO_IDX[self.type],
            self.world.COLOR_TO_IDX[self.color],
            self.state,
        )

    def render(self, img):
        c = self.world.COLORS[self.color]

        fill_coords(img, point_in_rect(0, 1, 0, 1), c)



from matplotlib import animation
import matplotlib.pyplot as plt
from wildfire_environment.core.constants import (
    STATE_IDX_TO_COLOR_WILDFIRE,
    TILE_PIXELS,
)
from wildfire_environment.utils.rendering import (
    fill_coords,
    point_in_circle,
    point_in_rect,
)
from wildfire_environment.core.agent import Agent
from wildfire_environment.core.grid import Grid
from wildfire_environment.core.world import WorldT


def save_frames_as_gif(
    frames, path="./", filename="env-render-episode", ep=0, fps=60, dpi=72
):
    """Save a list of frames as a gif.

    Parameters
    ----------
    frames : ndarray
        list of frames to be saved as gif. Each frame is an RGB image of the environment at a time step.
    path : str, optional
       location to save the gif, by default "./"
    filename : str, optional
        name of the gif, by default "env-render-episode-". The episode number is appended to the filename. Thus, entire path of saved gif is "path/filename"+"ep.gif"
    ep : int, optional
        the episode number during which the frames were recorded, by default 0. For example, it is useful if the frames are saved during RL agent training to identify the episode number during which the frames were recorded.
    fps : int, optional
        frequency at which consecutive images or frames are displayed in the gif, by default 60
    dpi : int, optional
        output resolution of gif, by default 72
    """
    filename = filename + "-" + str(ep) + ".gif"
    plt.figure(figsize=(frames[0].shape[1] / 72.0, frames[0].shape[0] / 72.0), dpi=dpi)

    patch = plt.imshow(frames[0])
    plt.axis("off")

    def animate(i):
        patch.set_data(frames[i])

    anim = animation.FuncAnimation(plt.gcf(), animate, frames=len(frames), interval=50)
    anim.save(path + filename, writer="imagemagick", fps=fps)
    plt.close()


def render_agent_tiles(
    img,
    agent: Agent,
    helper_grid: Grid,
    world: WorldT,
    x_min=None,
    y_min=None,
    x_max=None,
    y_max=None,
    colors=None,
):
    """
    Re-render the tile containing given agent to add background color corresponding to the state of tree in that cell.

    Parameters
    ----------
    img : NDArray
        image of wildfire grid with tree missing at agent location
    agent : Agent
        agent located in the tile to be re-rendered
    helper_grid : Grid
        grid containing only trees and no agents. Used to get the state of tree in the cell containing the agent
    world : WorldT
        wildfire world
    x_min : list[int], optional
        list of x-coordinates of the left boundary of selfish regions
    y_min : list[int], optional
        list of y-coordinates of the top boundary of selfish regions
    x_max : list[int], optional
        list of x-coordinates of the right boundary of selfish regions
    y_max : list[int], optional
        list of y-coordinates of the bottom boundary of selfish regions
    colors : list, optional
        list of colors of the boundaries of selfish regions. Boundary color is same as the color of the corresponding selfish agent

    Returns
    -------
    img : NDArray
        image with all trees and agents rendered
    """

    pos = agent.pos
    o = helper_grid.get(*pos)
    s = o.state

    tile_size = TILE_PIXELS
    ymin = pos[1] * tile_size
    ymax = (pos[1] + 1) * tile_size
    xmin = pos[0] * tile_size
    xmax = (pos[0] + 1) * tile_size

    tree_color = world.COLORS[STATE_IDX_TO_COLOR_WILDFIRE[s]]
    fill_coords(
        img[ymin:ymax, xmin:xmax, :],
        point_in_circle(0.5, 0.5, 0.25),
        world.COLORS[agent.color],
        bg_color=tree_color,
    )
    i, j = pos
    changed_left_boundary = False
    changed_top_boundary = False
    if colors is not None:
        for index, color in enumerate(colors):
            # check if object is located adjacent to the top boundary of selfish region
            if j == y_min[index]:
                if x_min[index] <= i <= x_max[index]:
                    changed_top_boundary = True
                    fill_coords(
                        img[ymin:ymax, xmin:xmax, :],
                        point_in_rect(0, 1, 0, 0.093),
                        color,
                    )
            # check if object is located adjacent to the left boundary of selfish region
            if i == x_min[index]:
                if y_min[index] <= j <= y_max[index]:
                    changed_left_boundary = True
                    fill_coords(
                        img[ymin:ymax, xmin:xmax, :],
                        point_in_rect(0, 0.093, 0, 1),
                        color,
                    )
            # check if object is located adjacent to the bottom boundary of selfish region
            if j == y_max[index] + 1:
                if x_min[index] <= i <= x_max[index]:
                    changed_top_boundary = True
                    fill_coords(
                        img[ymin:ymax, xmin:xmax, :],
                        point_in_rect(0, 1, 0, 0.093),
                        color,
                    )
            # check if object is located adjacent to the right boundary of selfish region
            if i == x_max[index] + 1:
                if y_min[index] <= j <= y_max[index]:
                    changed_left_boundary = True
                    fill_coords(
                        img[ymin:ymax, xmin:xmax, :],
                        point_in_rect(0, 0.093, 0, 1),
                        color,
                    )
    # use default boundary color if cell is not on boundary of selfish region
    if not changed_left_boundary:
        fill_coords(
            img[ymin:ymax, xmin:xmax, :], point_in_rect(0, 0.031, 0, 1), (100, 100, 100)
        )
    if not changed_top_boundary:
        fill_coords(
            img[ymin:ymax, xmin:xmax, :], point_in_rect(0, 1, 0, 0.031), (100, 100, 100)
        )

    return img


def get_initial_fire_coordinates(x, y, grid_size, n, without_wall_coordinates=False):
    """Generate the coordinates of trees on fire in a uniformly randomly located square fire region of size n by n.

    Parameters:
    ----------
        x : int
            the x-coordinate of the center cell of the fire region if n is odd, or the x-coordinate of the top-left corner cell of the fire region if n is even
        y : int
            the y-coordinate of the center cell of the fire region if n is odd, or the y-coordinate of the top-left corner cell of the fire region if n is even
        grid_size : int
            the side of the square grid
        n : int
            the side of the square fire region
        without_wall_coordinates : bool, optional
            if True, the coordinates (x, y) are assumed to be given in, and output coordinates are returned in, grid without wall coordinates.

    Returns:
    -------
    coordinates : list(tuple(int, int))
        a list of tuples, where each tuple represents the position coordinates of a tree on fire in the fire region.
    """

    if without_wall_coordinates:
        # convert grid without wall coordinates to grid with wall coordinates
        x += 1
        y += 1

    if n % 2 == 0:
        # side of the fire region is an even number
        coordinates = []
        # loop through the positions of cells in the fire region. The top-left corner cell is (x, y)
        for i in range(x, x + n):
            for j in range(y, y + n):
                coordinates.append((i, j))
        return coordinates
    else:
        # side of the fire region is an odd number
        # offset is the distance from the center cell to the edge cell of a fire region of size n by n
        offset = int((n - 1) / 2)

        # determine range of x and y coordinates lying within the fire region. The center cell is (x, y).
        start_x = max(1, x - offset)
        end_x = min(grid_size, x + offset)
        start_y = max(1, y - offset)
        end_y = min(grid_size, y + offset)

        coordinates = []
        # loop through the positions of cells in the fire region
        for x in range(start_x, end_x + 1):
            for y in range(start_y, end_y + 1):
                if without_wall_coordinates:
                    coordinates.append((x - 1, y - 1))
                else:
                    coordinates.append((x, y))
        return coordinates



from copy import deepcopy
import numpy as np
from wildfire_environment.core.world import WorldT
from wildfire_environment.utils.rendering import *
from wildfire_environment.core.object import WorldObj, Wall, WorldObjT
from wildfire_environment.core.constants import TILE_PIXELS


class Grid:
    """
    Represent a grid and operations on it
    """

    # Static cache of pre-renderer tiles
    tile_cache = {}

    def __init__(self, width: int, height: int, world: WorldT):
        """Create a grid of a given width and height in given world

        Parameters
        ----------
        width : int
            width of the grid
        height : int
            height of the grid
        world : WorldT
            world object in which the grid is situated
        """
        assert width >= 3
        assert height >= 3

        self.width = width
        self.height = height
        self.world = world

        self.grid = [None] * width * height

    def __contains__(self, key) -> bool:
        if isinstance(key, WorldObj):
            for e in self.grid:
                if e is key:
                    return True
        elif isinstance(key, tuple):
            for e in self.grid:
                if e is None:
                    continue
                if (e.color, e.type) == key:
                    return True
                if key[0] is None and key[1] == e.type:
                    return True
        return False

    def __eq__(self, other: "Grid") -> bool:
        grid1 = self.encode()
        grid2 = other.encode()
        return np.array_equal(grid2, grid1)

    def __ne__(self, other: "Grid") -> bool:
        return not self == other

    def copy(self) -> "Grid":
        """Create a deep copy of the grid

        Returns
        -------
        Grid
            deep copy of the grid
        """

        return deepcopy(self)

    def set(self, i: int, j: int, v) -> None:
        """Insert the given object at the given position in the grid

        Parameters
        ----------
        i : int
            x-coordinate of the position
        j : int
            y-coordinate of the position
        v : WorldObjT | None
            object to be inserted
        """
        assert i >= 0 and i < self.width
        assert j >= 0 and j < self.height
        self.grid[j * self.width + i] = v

    def get(self, i: int, j: int):
        """Get the object at the given position in the grid

        Parameters
        ----------
        i : int
            x-coordinate of the position
        j : int
            y-coordinate of the position

        Returns
        -------
        WorldObjT | None
            object at the given position in the grid
        """
        assert i >= 0 and i < self.width
        assert j >= 0 and j < self.height
        return self.grid[j * self.width + i]

    def horz_wall(
        self,
        x: int,
        y: int,
        length=None,
        obj_type=Wall,
    ) -> None:
        """Create a horizontal wall starting from given point (x, y) and of given length.

        Parameters
        ----------
        x : int
            x-coordinate of the starting point
        y : int
            y-coordinate of the starting point
        length : int | None, optional
            length of the wall, by default None
        obj_type : Type[WorldObjT], optional
            type of object to be inserted, by default Wall
        """
        if length is None:
            length = self.width - x
        assert length is not None
        for i in range(0, length):
            wall_obj = obj_type(self.world)
            wall_obj.pos = (x + i, y)
            self.set(x + i, y, wall_obj)

    def vert_wall(
        self,
        x: int,
        y: int,
        length=None,
        obj_type=Wall,
    ):
        """Create a vertical wall starting from given point (x, y) and of given length.

        Parameters
        ----------
        x : int
            x-coordinate of the starting point
        y : int
            y-coordinate of the starting point
        length : int | None, optional
            length of the wall, by default None
        obj_type : Type[WorldObjT], optional
            type of object to be inserted, by default Wall
        """
        if length is None:
            length = self.height - y
        for j in range(0, length):
            wall_obj = obj_type(self.world)
            wall_obj.pos = (x, y + j)
            self.set(x, y + j, wall_obj)

    def wall_rect(self, x: int, y: int, w: int, h: int) -> None:
        """Create a rectangle of walls starting from given point (x, y) and of given width and height.

        Parameters
        ----------
        x : int
            x-coordinate of the starting point
        y : int
            y-coordinate of the starting point
        w : int
            width of the rectangle
        h : int
            height of the rectangle
        """
        self.horz_wall(x, y, w)
        self.horz_wall(x, y + h - 1, w)
        self.vert_wall(x, y, h)
        self.vert_wall(x + w - 1, y, h)

    def rotate_left(self) -> "Grid":
        """
        Rotate the grid to the left (counter-clockwise)
        """

        grid = Grid(self.height, self.width, self.world)

        for i in range(self.width):
            for j in range(self.height):
                v = self.get(i, j)
                grid.set(j, grid.height - 1 - i, v)

        return grid

    def slice(self, topX, topY, width, height):
        """
        Get a subset of the grid. The subset is a rectangle of size width x height whose top-left corner is at (topX, topY).

        Parameters
        ----------
        topX : int
            x-coordinate of the top-left corner of the subset of grid
        topY : int
            y-coordinate of the top-left corner of the subset of grid
        width : int
            width of the subset of grid
        height : int
            height of the subset of grid
        """

        grid = Grid(width, height, self.world)

        for j in range(0, height):
            for i in range(0, width):
                x = topX + i
                y = topY + j

                if x >= 0 and x < self.width and y >= 0 and y < self.height:
                    v = self.get(x, y)
                else:
                    v = Wall(self.world)

                grid.set(i, j, v)

        return grid

    @classmethod
    def render_tile(
        cls,
        world: WorldT,
        obj,
        highlights=None,
        tile_size: int = TILE_PIXELS,
        subdivs: int = 3,
        cache: bool = True,
        cell_location: int = 0,
        selfish_boundary_color=(100, 100, 100),
    ):
        """
        Render a tile and cache the result

        Parameters
        ----------
        world : WorldT
            world object in which the grid is situated
        obj : WorldObjT | None
            object to be rendered
        highlights : list[bool], optional
            list of booleans indicating whether to highlight the tile, by default None
        tile_size : int, optional
            size of the tile, by default TILE_PIXELS
        subdivs : int, optional
            number of subdivisions to use for downsampling image, by default 3
        cache : bool, optional
            whether to cache the rendered tile, by default True
        cell_location : int, optional
            determine if the cell is located adjacent to a selfish region boundary, by default 0. Only applicable for wildfire environment.
        selfish_boundary_color : tuple[int, int, int], optional
            color of the selfish region boundary, by default (100, 100, 100). Only applicable for wildfire environment.

        """
        # Key for caching
        key = (*highlights, tile_size)
        key = obj.encode() + key if obj else key
        if cell_location != 0:
            key = (key, (cell_location, selfish_boundary_color.tobytes()))

        # Return the cached tile if it exists
        if key in cls.tile_cache:
            return cls.tile_cache[key]

        img = np.zeros(
            shape=(tile_size * subdivs, tile_size * subdivs, 3), dtype=np.uint8
        )

        # render the object
        if obj is not None:
            obj.render(img)

        # create grid lines around object (specifically the top and left boundaries for each cell)
        changed_left_boundary = False
        changed_top_boundary = False
        if cell_location == 1 or cell_location == 3:
            changed_top_boundary = True
            fill_coords(img, point_in_rect(0, 1, 0, 0.093), selfish_boundary_color)
        if cell_location == 2 or cell_location == 3:
            changed_left_boundary = True
            fill_coords(img, point_in_rect(0, 0.093, 0, 1), selfish_boundary_color)

        # use default boundary color if cell is not on boundary of selfish region
        if not changed_left_boundary:
            fill_coords(img, point_in_rect(0, 0.031, 0, 1), (100, 100, 100))
        if not changed_top_boundary:
            fill_coords(img, point_in_rect(0, 1, 0, 0.031), (100, 100, 100))

        # Highlight the cell  if needed
        if len(highlights) > 0:
            for h in highlights:
                highlight_img(
                    img,
                    color=world.COLORS[world.IDX_TO_COLOR[h % len(world.IDX_TO_COLOR)]],
                )

        # Downsample the image to perform supersampling/anti-aliasing
        img = downsample(img, subdivs)

        # Cache the rendered tile
        if cache:
            cls.tile_cache[key] = img
        else:
            pass

        return img

    def render(
        self,
        tile_size,
        highlight_masks=None,
        uncached_object_types=None,
        x_min=None,
        y_min=None,
        x_max=None,
        y_max=None,
        colors=None,
    ):
        """
        Render this grid at a given scale

        Parameters
        ----------
        tile_size : int
            size of the tile
        highlight_masks : list[bool], optional
            list of booleans indicating whether to highlight the tile, by default None
        uncached_object_types : list[str], optional
            list of object types that should not be cached, by default None
        x_min : list[int], optional
            list of x-coordinates of the left boundary of selfish regions, by default None. Only applicable for wildfire environment.
        y_min : list[int], optional
            list of y-coordinates of the top boundary of selfish regions, by default None. Only applicable for wildfire environment.
        x_max : list[int], optional
            list of x-coordinates of the right boundary of selfish regions, by default None. Only applicable for wildfire environment.
        y_max : list[int], optional
            list of y-coordinates of the bottom boundary of selfish regions, by default None. Only applicable for wildfire environment.
        colors : list[tuple[int,int,int]], optional
            list of colors to use for selfish region boundaries, by default None. Only applicable for wildfire environment.
        """

        # Compute the total grid size
        width_px = self.width * tile_size
        height_px = self.height * tile_size

        img = np.zeros(shape=(height_px, width_px, 3), dtype=np.uint8)

        # Render the grid
        for j in range(0, self.height):
            for i in range(0, self.width):
                cell = self.get(i, j)
                cache: bool = True
                if cell is not None and cell.type in uncached_object_types:
                    cache = False
                if x_min is not None:
                    # determine if the cell is located adjacent to a selfish region boundary
                    cell_location = 0
                    selfish_boundary_color = (100, 100, 100)
                    for index, color in enumerate(colors):
                        # check if object is located adjacent to the top boundary of selfish region
                        if j == y_min[index]:
                            if x_min[index] <= i <= x_max[index]:
                                cell_location = 1
                                selfish_boundary_color = color
                        # check if object is located adjacent to the left boundary of selfish region
                        if i == x_min[index]:
                            if y_min[index] <= j <= y_max[index]:
                                cell_location += 2
                                selfish_boundary_color = color
                        # check if object is located adjacent to the bottom boundary of selfish region
                        if j == y_max[index] + 1:
                            if x_min[index] <= i <= x_max[index]:
                                cell_location = 1
                                selfish_boundary_color = color
                        # check if object is located adjacent to the right boundary of selfish region
                        if i == x_max[index] + 1:
                            if y_min[index] <= j <= y_max[index]:
                                cell_location = 2
                                selfish_boundary_color = color
                    # render the tile
                    tile_img = Grid.render_tile(
                        self.world,
                        cell,
                        highlights=(
                            [] if highlight_masks is None else highlight_masks[i, j]
                        ),
                        tile_size=tile_size,
                        cache=cache,
                        cell_location=cell_location,
                        selfish_boundary_color=selfish_boundary_color,
                    )
                else:
                    # render the tile without selfish region boundary
                    tile_img = Grid.render_tile(
                        self.world,
                        cell,
                        highlights=(
                            [] if highlight_masks is None else highlight_masks[i, j]
                        ),
                        tile_size=tile_size,
                        cache=cache,
                    )

                ymin = j * tile_size
                ymax = (j + 1) * tile_size
                xmin = i * tile_size
                xmax = (i + 1) * tile_size
                img[ymin:ymax, xmin:xmax, :] = tile_img

        return img

    def encode(self, vis_mask=None):
        """
        Produce a compact numpy encoding of the grid

        Parameters
        ----------
        vis_mask : np.ndarray[bool] | None, optional
            mask specifying visible regions of grid, by default None

        Returns
        -------
        np.ndarray
            compact numpy encoding of the grid
        """

        if vis_mask is None:
            vis_mask = np.ones((self.width, self.height), dtype=bool)

        array = np.zeros(
            (self.width, self.height, self.world.encode_dim), dtype="uint8"
        )

        for i in range(self.width):
            for j in range(self.height):
                if vis_mask[i, j]:
                    v = self.get(i, j)

                    if v is None:
                        array[i, j, 0] = self.world.OBJECT_TO_IDX["empty"]
                        array[i, j, 1] = 0
                        array[i, j, 2] = 0
                        if self.world.encode_dim > 3:
                            array[i, j, 3] = 0
                            array[i, j, 4] = 0
                            array[i, j, 5] = 0

                    else:
                        array[i, j, :] = v.encode(self.world)

        return array

    def encode_for_agents(self, agent_pos, vis_mask=None):
        """
        Produce a compact numpy encoding of the grid

        Parameters
        ----------
        agent_pos : tuple[int, int]
            position of the agent
        vis_mask : np.ndarray[bool] | None, optional
            mask specifying visible regions of grid, by default None

        Returns
        -------
        np.ndarray
            compact numpy encoding of the grid
        """
        if vis_mask is None:
            vis_mask = np.ones((self.width, self.height), dtype=bool)

        array = np.zeros(
            (self.width, self.height, self.world.encode_dim), dtype="uint8"
        )

        for i in range(self.width):
            for j in range(self.height):
                if vis_mask[i, j]:
                    v = self.get(i, j)

                    if v is None:
                        array[i, j, 0] = self.world.OBJECT_TO_IDX["empty"]
                        array[i, j, 1] = 0
                        array[i, j, 2] = 0
                        if self.world.encode_dim > 3:
                            array[i, j, 3] = 0
                            array[i, j, 4] = 0
                            array[i, j, 5] = 0

                    else:
                        array[i, j, :] = v.encode(
                            current_agent=np.array_equal(agent_pos, (i, j))
                        )

        return array

    def process_vis(grid, agent_pos):
        """Returns a mask of the visible cells in the grid

        Parameters
        ----------
        grid : Grid

        agent_pos : tuple[int, int]
            position of the agent

        Returns
        -------
        np.ndarray[bool]
            mask of the visible cells in the grid
        """
        mask = np.zeros(shape=(grid.width, grid.height), dtype=bool)

        mask[agent_pos[0], agent_pos[1]] = True

        for j in reversed(range(0, grid.height)):
            for i in range(0, grid.width - 1):
                if not mask[i, j]:
                    continue

                cell = grid.get(i, j)
                if cell and not cell.see_behind():
                    continue

                mask[i + 1, j] = True
                if j > 0:
                    mask[i + 1, j - 1] = True
                    mask[i, j - 1] = True

            for i in reversed(range(1, grid.width)):
                if not mask[i, j]:
                    continue

                cell = grid.get(i, j)
                if cell and not cell.see_behind():
                    continue

                mask[i - 1, j] = True
                if j > 0:
                    mask[i - 1, j - 1] = True
                    mask[i, j - 1] = True

        for j in range(0, grid.height):
            for i in range(0, grid.width):
                if not mask[i, j]:
                    grid.set(i, j, None)

        return mask



import math
from typing import TypeVar
import numpy as np
import gym
from gym import spaces
from gym.utils import seeding
import random
from wildfire_environment.core.grid import Grid
from wildfire_environment.core.object import WorldObjT
from wildfire_environment.core.world import DefaultWorld, World, WorldT
from wildfire_environment.core.agent import (
    AgentT,
    DefaultActions,
)
from wildfire_environment.utils.rendering import *
from wildfire_environment.utils.window import Window
from wildfire_environment.core.constants import *


MultiGridEnvT = TypeVar("MultiGridEnvT", bound="MultiGridEnv")


class MultiGridEnv(gym.Env):
    """
    2D grid world game environment
    """

    metadata = {"render_modes": ["human", "rgb_array"], "video.frames_per_second": 10}

    def __init__(
        self,
        agents,
        grid_size=None,
        width=None,
        height=None,
        max_steps: int = 100,
        see_through_walls: bool = False,
        partial_obs: bool = False,
        agent_view_size: int = 7,
        actions_set=DefaultActions,
        world: WorldT = DefaultWorld,
        render_mode="rgb_array",
        uncached_object_types=[],
    ):
        self.agents = agents
        self.render_mode = render_mode
        self.uncahed_object_types = uncached_object_types
        # Does the agents have partial or full observation?
        self.partial_obs = partial_obs
        self.agent_view_size = agent_view_size

        # Can't set both grid_size and width/height
        if grid_size:
            assert width == None and height == None
            width = grid_size
            height = grid_size
        else:
            assert width != None and height != None

        self.width: int = width
        self.height: int = height

        # Action enumeration for this environment
        self.actions = actions_set

        # Actions are discrete integer values
        self.action_space = spaces.Discrete(len(self.actions))

        self.world = world

        self.observation_space = self._set_observation_space()

        if self.observation_space is spaces.Box:
            self.ob_dim = np.prod(self.observation_space.shape)
        else:
            pass
        self.ac_dim = self.action_space.n

        # Range of possible rewards
        self.reward_range = (0, 1)

        # Window to use for human rendering mode
        self.window = None

        # Environment configuration
        self.max_steps = max_steps
        self.see_through_walls = see_through_walls

        # Initialize the RNG

        # Define the empty grid. _gen_grid is supposed to fill this up
        self.grid = Grid(width, height, world)

    def _set_observation_space(self):
        if self.partial_obs:
            observation_space = spaces.Box(
                low=0,
                high=255,
                shape=(
                    self.agent_view_size,
                    self.agent_view_size,
                    self.world.encode_dim,
                ),
                dtype="uint8",
            )

        else:
            observation_space = spaces.Box(
                low=0,
                high=255,
                shape=(self.width, self.height, self.world.encode_dim),
                dtype="uint8",
            )

        return observation_space

    def reset(self, seed=None, state=None):
        if seed is not None:
            self._np_random, seed = seeding.np_random(seed)
        # Generate a new random grid at the start of each episode
        # To keep the same grid for each episode, call env.seed() with
        # the same seed before calling env.reset()
        if state is not None:
            self._gen_grid(self.width, self.height, state)
        else:
            self._gen_grid(self.width, self.height)

        # These fields should be defined by _gen_grid
        for a in self.agents:
            assert a.pos is not None
            assert a.dir is not None

        # Item picked up, being carried, initially nothing
        for a in self.agents:
            a.carrying = None

        # Step count since episode start
        self.step_count = 0

        # Return first observation
        if self.partial_obs:
            obs = self.gen_obs()
        else:
            obs = [
                self.grid.encode_for_agents(agent_pos=self.agents[i].pos)
                for i in range(len(self.agents))
            ]
        obs = [self.world.normalize_obs * ob for ob in obs]
        return obs

    @property
    def steps_remaining(self):
        return self.max_steps - self.step_count

    def __str__(self):
        """
        Produce a pretty string of the environment's grid along with the agent.
        A grid cell is represented by 2-character string, the first one for
        the object and the second one for the color.
        """

        str = ""

        for j in range(self.grid.height):
            for i in range(self.grid.width):
                # if i == self.agent_pos[0] and j == self.agent_pos[1]:
                # str += 2 * AGENT_DIR_TO_STR[self.agent_dir]
                # continue

                c = self.grid.get(i, j)

                if c == None:
                    str += "  "
                    continue

                if c.type == "door":
                    if c.is_open:
                        str += "__"
                    elif c.is_locked:
                        str += "L" + c.color[0].upper()
                    else:
                        str += "D" + c.color[0].upper()
                    continue

                str += OBJECT_TO_STR[c.type] + c.color[0].upper()

            if j < self.grid.height - 1:
                str += "\n"

        return str

    def _gen_grid(self, width, height):
        self.grid = Grid(width, height, self.world)
        assert False, "_gen_grid needs to be implemented by each environment"

    def _handle_pickup(self, i, rewards, fwd_pos, fwd_cell):
        pass

    def _handle_build(self, i, rewards, fwd_pos, fwd_cell):
        pass

    def _handle_drop(self, i, rewards, fwd_pos, fwd_cell):
        pass

    def _handle_special_moves(self, i, rewards, fwd_pos, fwd_cell):
        pass

    def _handle_switch(self, i, rewards, fwd_pos, fwd_cell):
        pass

    def _reward(self, current_agent, rewards, reward=1):
        """
        Compute the reward to be given upon success
        """
        rewards[current_agent] += reward - 0.9 * (self.step_count / self.max_steps)
        return rewards

    def _rand_int(self, low, high):
        """
        Generate random integer in [low,high[
        """

        return random.randint(low, high)

    def _rand_float(self, low, high):
        """
        Generate random float in [low,high[
        """

        return self.np_random.uniform(low, high)

    def _rand_bool(self):
        """
        Generate random boolean value
        """

        return self.np_random.randint(0, 2) == 0

    def _rand_elem(self, iterable):
        """
        Pick a random element in a list
        """

        lst = list(iterable)
        idx = self._rand_int(0, len(lst))
        return lst[idx]

    def _rand_subset(self, iterable, num_elems):
        """
        Sample a random subset of distinct elements of a list
        """

        lst = list(iterable)
        assert num_elems <= len(lst)

        out = []

        while len(out) < num_elems:
            elem = self._rand_elem(lst)
            lst.remove(elem)
            out.append(elem)

        return out

    def _rand_pos(self, xLow, xHigh, yLow, yHigh):
        """
        Generate a random (x,y) position tuple
        """

        return (
            self.np_random.randint(xLow, xHigh),
            self.np_random.randint(yLow, yHigh),
        )

    def place_obj(
        self,
        obj: WorldObjT,
        top=None,
        size=None,
        reject_fn=None,
        max_tries: float = math.inf,
    ):
        """
        Place an object at an empty position in the grid

        :param top: top-left position of the rectangle where to place
        :param size: size of the rectangle where to place
        :param reject_fn: function to filter out potential positions
        """

        if top is None:
            top = (0, 0)
        else:
            top = (max(top[0], 0), max(top[1], 0))

        if size is None:
            size = (self.grid.width, self.grid.height)

        num_tries = 0

        while True:
            # This is to handle with rare cases where rejection sampling
            # gets stuck in an infinite loop
            if num_tries > max_tries:
                raise RecursionError("rejection sampling failed in place_obj")

            num_tries += 1

            pos = np.array(
                (
                    self._rand_int(top[0], min(top[0] + size[0], self.grid.width - 1)),
                    self._rand_int(top[1], min(top[1] + size[1], self.grid.height - 1)),
                )
            )

            # Don't place the object on top of another object
            if self.grid.get(*pos) != None:
                continue

            # Check if there is a filtering criterion
            if reject_fn and reject_fn(self, pos):
                continue

            break

        self.grid.set(*pos, obj)

        if obj is not None:
            obj.init_pos = pos
            obj.pos = pos

        return pos

    def put_obj(self, obj: WorldObjT, i: int, j: int):
        """
        Put an object at a specific position in the grid
        """

        self.grid.set(i, j, obj)
        obj.init_pos = (i, j)
        obj.pos = (i, j)

    def place_agent(
        self,
        agent: AgentT,
        pos=None,
        top=None,
        size=None,
        rand_dir: bool = False,
        max_tries: float = math.inf,
    ):
        """
        Set the agent's starting point at an empty position in the grid
        """
        if pos is not None:
            agent.pos = pos
            self.put_obj(agent, i=pos[0], j=pos[1])
        else:
            agent.pos = None
            pos = self.place_obj(agent, top, size, max_tries=max_tries)
            agent.pos = pos
            agent.init_pos = pos

        if rand_dir:
            agent.dir = self._rand_int(0, 3)
        else:
            agent.dir = 3

        agent.init_dir = agent.dir

        return pos

    def agent_sees(self, a, x, y):
        """
        Check if a non-empty grid position is visible to the agent
        """

        coordinates = a.relative_coords(x, y)
        if coordinates is None:
            return False
        vx, vy = coordinates

        obs = self.gen_obs()
        obs_grid, _ = self.grid.decode(obs["image"])
        obs_cell = obs_grid.get(vx, vy)
        world_cell = self.grid.get(x, y)

        return obs_cell is not None and obs_cell.type == world_cell.type

    def step(self, actions):
        self.step_count += 1

        order = np.random.permutation(len(actions))

        rewards = np.zeros(len(actions))
        done = False

        for i in order:
            if (
                self.agents[i].terminated
                or self.agents[i].paused
                or not self.agents[i].started
                or actions[i] == self.actions.still
            ):
                continue

            # Get the position in front of the agent
            fwd_pos = self.agents[i].front_pos

            # Get the contents of the cell in front of the agent
            fwd_cell = self.grid.get(*fwd_pos)

            # Rotate left
            if actions[i] == self.actions.left:
                self.agents[i].dir -= 1
                if self.agents[i].dir < 0:
                    self.agents[i].dir += 4

            # Rotate right
            elif actions[i] == self.actions.right:
                self.agents[i].dir = (self.agents[i].dir + 1) % 4

            # Move forward
            elif actions[i] == self.actions.forward:
                if fwd_cell is not None:
                    if fwd_cell.type == "goal":
                        done = True
                        rewards = self._reward(i, rewards, 1)
                    elif fwd_cell.type == "switch":
                        self._handle_switch(i, rewards, fwd_pos, fwd_cell)
                    elif fwd_cell.type == "ball":
                        rewards = self._handle_pickup(i, rewards, fwd_pos, fwd_cell)
                elif fwd_cell is None or fwd_cell.can_overlap():
                    self.grid.set(*fwd_pos, self.agents[i])
                    self.grid.set(*self.agents[i].pos, None)
                    self.agents[i].pos = fwd_pos
                self._handle_special_moves(i, rewards, fwd_pos, fwd_cell)

            elif "build" in self.actions.available and actions[i] == self.actions.build:
                self._handle_build(i, rewards, fwd_pos, fwd_cell)

            # Pick up an object
            elif actions[i] == self.actions.pickup:
                self._handle_pickup(i, rewards, fwd_pos, fwd_cell)

            # Drop an object
            elif actions[i] == self.actions.drop:
                self._handle_drop(i, rewards, fwd_pos, fwd_cell)

            # Toggle/activate an object
            elif actions[i] == self.actions.toggle:
                if fwd_cell:
                    fwd_cell.toggle(self, fwd_pos)

            # Done action (not used by default)
            elif actions[i] == self.actions.done:
                pass

            else:
                assert False, "unknown action"

        if self.step_count >= self.max_steps:
            done = True

        if self.partial_obs:
            obs = self.gen_obs()
        else:
            obs = [
                self.grid.encode_for_agents(
                    world=self.world, agent_pos=self.agents[i].pos
                )
                for i in range(len(actions))
            ]

        obs = [self.world.normalize_obs * ob for ob in obs]

        return obs, rewards, done, {}

    def gen_obs_grid(self):
        """
        Generate the sub-grid observed by the agents.
        This method also outputs a visibility mask telling us which grid
        cells the agents can actually see.
        """

        grids = []
        vis_masks = []

        for a in self.agents:
            topX, topY, botX, botY = a.get_view_exts()

            grid = self.grid.slice(self.world, topX, topY, a.view_size, a.view_size)

            for i in range(a.dir + 1):
                grid = grid.rotate_left()

            # Process occluders and visibility
            # Note that this incurs some performance cost
            if not self.see_through_walls:
                vis_mask = grid.process_vis(
                    agent_pos=(a.view_size // 2, a.view_size - 1)
                )
            else:
                vis_mask = np.ones(shape=(grid.width, grid.height), dtype=np.bool)

            grids.append(grid)
            vis_masks.append(vis_mask)

        return grids, vis_masks

    def gen_obs(self):
        """
        Generate the agent's view (partially observable, low-resolution encoding)
        """

        grids, vis_masks = self.gen_obs_grid()

        # Encode the partially observable view into a numpy array
        obs = [
            grid.encode_for_agents(
                self.world, [grid.width // 2, grid.height - 1], vis_mask
            )
            for grid, vis_mask in zip(grids, vis_masks)
        ]

        return obs

    def get_obs_render(self, obs, tile_size=TILE_PIXELS // 2):
        """
        Render an agent observation for visualization
        """

        grid, vis_mask = self.grid.decode(obs)

        # Render the whole grid
        img = grid.render(self.world, tile_size, highlight_mask=vis_mask)

        return img

    def render(self, close=False, highlight=False, tile_size=TILE_PIXELS):
        """
        Render the whole-grid human view
        """

        if close:
            if self.window:
                self.window.close()
            return

        if self.render_mode == "human" and not self.window:
            self.window = Window("gym_multigrid")
            self.window.show(block=False)

        if highlight:
            # Compute which cells are visible to the agent
            _, vis_masks = self.gen_obs_grid()

            highlight_masks = {
                (i, j): [] for i in range(self.width) for j in range(self.height)
            }

            for i, a in enumerate(self.agents):
                # Compute the world coordinates of the bottom-left corner
                # of the agent's view area
                f_vec = a.dir_vec
                r_vec = a.right_vec
                top_left = (
                    a.pos + f_vec * (a.view_size - 1) - r_vec * (a.view_size // 2)
                )

                # Mask of which cells to highlight

                # For each cell in the visibility mask
                for vis_j in range(0, a.view_size):
                    for vis_i in range(0, a.view_size):
                        # If this cell is not visible, don't highlight it
                        if not vis_masks[i][vis_i, vis_j]:
                            continue

                        # Compute the world coordinates of this cell
                        abs_i, abs_j = top_left - (f_vec * vis_j) + (r_vec * vis_i)

                        if abs_i < 0 or abs_i >= self.width:
                            continue
                        if abs_j < 0 or abs_j >= self.height:
                            continue

                        # Mark this cell to be highlighted
                        highlight_masks[abs_i, abs_j].append(i)

        # Render the whole grid
        img = self.grid.render(
            tile_size,
            highlight_masks=highlight_masks if highlight else None,
            uncached_object_types=self.uncahed_object_types,
        )

        if self.render_mode == "human":
            self.window.show_img(img)

        return img



import enum
from abc import abstractmethod, ABC
from typing import TypeVar, TypedDict, Type

AgentPolicyT = TypeVar("AgentPolicyT", bound="BaseAgentPolicy")


class BaseAgentPolicy(ABC):
    """
    Abstract class for CTF enemy policy
    """

    def __init__(self) -> None:
        super().__init__()
        self.name: str = "base"

    @abstractmethod
    def act(self, observation: TypedDict, actions: Type[enum.IntEnum]) -> int:
        ...
